import json
import sys
from textwrap import dedent

import pytest


def test_simple_run(tmpdir, runner):
    runner.write_with_general(
        dedent(
            """
    [pair my_pair]
    a = "my_a"
    b = "my_b"
    collections = null

    [storage my_a]
    type = "filesystem"
    path = "{0}/path_a/"
    fileext = ".txt"

    [storage my_b]
    type = "filesystem"
    path = "{0}/path_b/"
    fileext = ".txt"
    """
        ).format(str(tmpdir))
    )

    tmpdir.mkdir("path_a")
    tmpdir.mkdir("path_b")

    result = runner.invoke(["discover"])
    assert not result.exception

    result = runner.invoke(["sync"])
    assert not result.exception

    tmpdir.join("path_a/haha.txt").write("UID:haha")
    result = runner.invoke(["sync"])
    assert "Copying (uploading) item haha to my_b" in result.output
    assert tmpdir.join("path_b/haha.txt").read() == "UID:haha"


def test_sync_inexistant_pair(tmpdir, runner):
    runner.write_with_general("")

    result = runner.invoke(["sync", "foo"])
    assert result.exception
    assert "pair foo does not exist." in result.output.lower()


def test_debug_connections(tmpdir, runner):
    runner.write_with_general(
        dedent(
            """
    [pair my_pair]
    a = "my_a"
    b = "my_b"
    collections = null

    [storage my_a]
    type = "filesystem"
    path = "{0}/path_a/"
    fileext = ".txt"

    [storage my_b]
    type = "filesystem"
    path = "{0}/path_b/"
    fileext = ".txt"
    """
        ).format(str(tmpdir))
    )

    tmpdir.mkdir("path_a")
    tmpdir.mkdir("path_b")

    result = runner.invoke(["discover"])
    assert not result.exception

    result = runner.invoke(["-vdebug", "sync", "--max-workers=3"])
    assert "using 3 maximal workers" in result.output.lower()

    result = runner.invoke(["-vdebug", "sync"])
    assert "using 1 maximal workers" in result.output.lower()


def test_empty_storage(tmpdir, runner):
    runner.write_with_general(
        dedent(
            """
    [pair my_pair]
    a = "my_a"
    b = "my_b"
    collections = null

    [storage my_a]
    type = "filesystem"
    path = "{0}/path_a/"
    fileext = ".txt"

    [storage my_b]
    type = "filesystem"
    path = "{0}/path_b/"
    fileext = ".txt"
    """
        ).format(str(tmpdir))
    )

    tmpdir.mkdir("path_a")
    tmpdir.mkdir("path_b")

    result = runner.invoke(["discover"])
    assert not result.exception

    result = runner.invoke(["sync"])
    assert not result.exception

    tmpdir.join("path_a/haha.txt").write("UID:haha")
    result = runner.invoke(["sync"])
    assert not result.exception
    tmpdir.join("path_b/haha.txt").remove()
    result = runner.invoke(["sync"])
    lines = result.output.splitlines()
    assert lines[0] == "Syncing my_pair"
    assert lines[1].startswith(
        "error: my_pair: " 'Storage "my_b" was completely emptied.'
    )
    assert result.exception


def test_verbosity(tmpdir, runner):
    runner.write_with_general("")
    result = runner.invoke(["--verbosity=HAHA", "sync"])
    assert result.exception
    assert (
        'invalid value for "--verbosity"' in result.output.lower()
        or "invalid value for '--verbosity'" in result.output.lower()
    )


def test_collections_cache_invalidation(tmpdir, runner):
    foo = tmpdir.mkdir("foo")
    bar = tmpdir.mkdir("bar")
    for x in "abc":
        foo.mkdir(x)
        bar.mkdir(x)

    runner.write_with_general(
        dedent(
            """
    [storage foo]
    type = "filesystem"
    path = "{0}/foo/"
    fileext = ".txt"

    [storage bar]
    type = "filesystem"
    path = "{0}/bar/"
    fileext = ".txt"

    [pair foobar]
    a = "foo"
    b = "bar"
    collections = ["a", "b", "c"]
    """
        ).format(str(tmpdir))
    )

    foo.join("a/itemone.txt").write("UID:itemone")

    result = runner.invoke(["discover"])
    assert not result.exception

    result = runner.invoke(["sync"])
    assert not result.exception
    assert "detected change in config file" not in result.output.lower()

    rv = bar.join("a").listdir()
    assert len(rv) == 1
    assert rv[0].basename == "itemone.txt"

    runner.write_with_general(
        dedent(
            """
    [storage foo]
    type = "filesystem"
    path = "{0}/foo/"
    fileext = ".txt"

    [storage bar]
    type = "filesystem"
    path = "{0}/bar2/"
    fileext = ".txt"

    [pair foobar]
    a = "foo"
    b = "bar"
    collections = ["a", "b", "c"]
    """
        ).format(str(tmpdir))
    )

    for entry in tmpdir.join("status").listdir():
        if not str(entry).endswith(".collections"):
            entry.remove()
    bar2 = tmpdir.mkdir("bar2")
    for x in "abc":
        bar2.mkdir(x)
    result = runner.invoke(["sync"])
    assert "detected change in config file" in result.output.lower()
    assert result.exception

    result = runner.invoke(["discover"])
    assert not result.exception

    result = runner.invoke(["sync"])
    assert not result.exception

    rv = bar.join("a").listdir()
    rv2 = bar2.join("a").listdir()
    assert len(rv) == len(rv2) == 1
    assert rv[0].basename == rv2[0].basename == "itemone.txt"


def test_invalid_pairs_as_cli_arg(tmpdir, runner):
    runner.write_with_general(
        dedent(
            """
    [storage foo]
    type = "filesystem"
    path = "{0}/foo/"
    fileext = ".txt"

    [storage bar]
    type = "filesystem"
    path = "{0}/bar/"
    fileext = ".txt"

    [pair foobar]
    a = "foo"
    b = "bar"
    collections = ["a", "b", "c"]
    """
        ).format(str(tmpdir))
    )

    for base in ("foo", "bar"):
        base = tmpdir.mkdir(base)
        for c in "abc":
            base.mkdir(c)

    result = runner.invoke(["discover"])
    assert not result.exception

    result = runner.invoke(["sync", "foobar/d"])
    assert result.exception
    assert 'pair foobar: collection "d" not found' in result.output.lower()


def test_multiple_pairs(tmpdir, runner):
    def get_cfg():
        for name_a, name_b in ("foo", "bar"), ("bam", "baz"):
            yield dedent(
                """
            [pair {a}{b}]
            a = "{a}"
            b = "{b}"
            collections = null
            """
            ).format(a=name_a, b=name_b)

            for name in name_a, name_b:
                yield dedent(
                    """
                [storage {name}]
                type = "filesystem"
                path = "{path}"
                fileext = ".txt"
                """
                ).format(name=name, path=str(tmpdir.mkdir(name)))

    runner.write_with_general("".join(get_cfg()))

    result = runner.invoke(["discover"])
    assert not result.exception
    assert set(result.output.splitlines()) > {
        "Discovering collections for pair bambaz",
        "Discovering collections for pair foobar",
    }

    result = runner.invoke(["sync"])
    assert not result.exception
    assert set(result.output.splitlines()) == {
        "Syncing bambaz",
        "Syncing foobar",
    }


# XXX: https://github.com/pimutils/vdirsyncer/issues/617
@pytest.mark.skipif(sys.platform == "darwin", reason="This test inexplicably fails")
@pytest.mark.parametrize(
    "collections",
    [
        ("a", "A"),
        ("\ufffe",),
        ("Hello there!",),
        ("Österreich",),
        ("中国", "x1"),
        ("한글",),
        ("42a4ec99-b1c2-4859-b142-759112f2ca50",),
        ("فلسطين",),
    ],
)
def test_create_collections(collections, tmpdir, runner):

    runner.write_with_general(
        dedent(
            """
    [pair foobar]
    a = "foo"
    b = "bar"
    collections = {colls}

    [storage foo]
    type = "filesystem"
    path = "{base}/foo/"
    fileext = ".txt"

    [storage bar]
    type = "filesystem"
    path = "{base}/bar/"
    fileext = ".txt"
    """.format(
                base=str(tmpdir), colls=json.dumps(list(collections))
            )
        )
    )

    result = runner.invoke(["discover"], input="y\n" * 2 * (len(collections) + 1))
    assert not result.exception, result.output

    result = runner.invoke(["sync"] + ["foobar/" + x for x in collections])
    assert not result.exception, result.output

    assert {x.basename for x in tmpdir.join("foo").listdir()} == {
        x.basename for x in tmpdir.join("bar").listdir()
    }


def test_ident_conflict(tmpdir, runner):
    runner.write_with_general(
        dedent(
            """
    [pair foobar]
    a = "foo"
    b = "bar"
    collections = null

    [storage foo]
    type = "filesystem"
    path = "{base}/foo/"
    fileext = ".txt"

    [storage bar]
    type = "filesystem"
    path = "{base}/bar/"
    fileext = ".txt"
    """.format(
                base=str(tmpdir)
            )
        )
    )

    foo = tmpdir.mkdir("foo")
    tmpdir.mkdir("bar")

    foo.join("one.txt").write("UID:1")
    foo.join("two.txt").write("UID:1")
    foo.join("three.txt").write("UID:1")

    result = runner.invoke(["discover"])
    assert not result.exception

    result = runner.invoke(["sync"])
    assert result.exception
    assert (
        'error: foobar: Storage "foo" contains multiple items with the '
        "same UID or even content"
    ) in result.output
    assert (
        sorted(
            [
                "one.txt" in result.output,
                "two.txt" in result.output,
                "three.txt" in result.output,
            ]
        )
        == [False, True, True]
    )


@pytest.mark.parametrize(
    "existing,missing",
    [
        ("foo", "bar"),
        ("bar", "foo"),
    ],
)
def test_unknown_storage(tmpdir, runner, existing, missing):
    runner.write_with_general(
        dedent(
            """
    [pair foobar]
    a = "foo"
    b = "bar"
    collections = null

    [storage {existing}]
    type = "filesystem"
    path = "{base}/{existing}/"
    fileext = ".txt"
    """.format(
                base=str(tmpdir), existing=existing
            )
        )
    )

    tmpdir.mkdir(existing)

    result = runner.invoke(["discover"])
    assert result.exception

    assert (
        "Storage '{missing}' not found. "
        "These are the configured storages: ['{existing}']".format(
            missing=missing, existing=existing
        )
    ) in result.output


@pytest.mark.parametrize("cmd", ["sync", "metasync"])
def test_no_configured_pairs(tmpdir, runner, cmd):
    runner.write_with_general("")

    result = runner.invoke([cmd])
    assert result.output == "critical: Nothing to do.\n"
    assert result.exception.code == 5


@pytest.mark.parametrize(
    "resolution,expect_foo,expect_bar",
    [(["command", "cp"], "UID:lol\nfööcontent", "UID:lol\nfööcontent")],
)
def test_conflict_resolution(tmpdir, runner, resolution, expect_foo, expect_bar):
    runner.write_with_general(
        dedent(
            """
    [pair foobar]
    a = "foo"
    b = "bar"
    collections = null
    conflict_resolution = {val}

    [storage foo]
    type = "filesystem"
    fileext = ".txt"
    path = "{base}/foo"

    [storage bar]
    type = "filesystem"
    fileext = ".txt"
    path = "{base}/bar"
    """.format(
                base=str(tmpdir), val=json.dumps(resolution)
            )
        )
    )

    foo = tmpdir.join("foo")
    bar = tmpdir.join("bar")
    fooitem = foo.join("lol.txt").ensure()
    fooitem.write("UID:lol\nfööcontent")
    baritem = bar.join("lol.txt").ensure()
    baritem.write("UID:lol\nbööcontent")

    r = runner.invoke(["discover"])
    assert not r.exception

    r = runner.invoke(["sync"])
    assert not r.exception

    assert fooitem.read() == expect_foo
    assert baritem.read() == expect_bar


@pytest.mark.parametrize("partial_sync", ["error", "ignore", "revert", None])
def test_partial_sync(tmpdir, runner, partial_sync):
    runner.write_with_general(
        dedent(
            """
    [pair foobar]
    a = "foo"
    b = "bar"
    collections = null
    {partial_sync}

    [storage foo]
    type = "filesystem"
    fileext = ".txt"
    path = "{base}/foo"

    [storage bar]
    type = "filesystem"
    read_only = true
    fileext = ".txt"
    path = "{base}/bar"
    """.format(
                partial_sync=(
                    f'partial_sync = "{partial_sync}"\n' if partial_sync else ""
                ),
                base=str(tmpdir),
            )
        )
    )

    foo = tmpdir.mkdir("foo")
    bar = tmpdir.mkdir("bar")

    foo.join("other.txt").write("UID:other")
    bar.join("other.txt").write("UID:other")

    baritem = bar.join("lol.txt")
    baritem.write("UID:lol")

    r = runner.invoke(["discover"])
    assert not r.exception

    r = runner.invoke(["sync"])
    assert not r.exception

    fooitem = foo.join("lol.txt")
    fooitem.remove()

    r = runner.invoke(["sync"])

    if partial_sync == "error":
        assert r.exception
        assert "Attempted change" in r.output
    elif partial_sync == "ignore":
        assert baritem.exists()
        r = runner.invoke(["sync"])
        assert not r.exception
        assert baritem.exists()
    else:
        assert baritem.exists()
        r = runner.invoke(["sync"])
        assert not r.exception
        assert baritem.exists()
        assert fooitem.exists()


def test_fetch_only_necessary_params(tmpdir, runner):
    fetched_file = tmpdir.join("fetched_flag")
    fetch_script = tmpdir.join("fetch_script")
    fetch_script.write(
        dedent(
            """
    set -e
    touch "{}"
    echo ".txt"
    """.format(
                str(fetched_file)
            )
        )
    )

    runner.write_with_general(
        dedent(
            """
    [pair foobar]
    a = "foo"
    b = "bar"
    collections = null

    [pair bambar]
    a = "bam"
    b = "bar"
    collections = null

    [storage foo]
    type = "filesystem"
    path = "{path}"
    fileext = ".txt"

    [storage bar]
    type = "filesystem"
    path = "{path}"
    fileext = ".txt"

    [storage bam]
    type = "filesystem"
    path = "{path}"
    fileext.fetch = ["command", "sh", "{script}"]
    """.format(
                path=str(tmpdir.mkdir("bogus")), script=str(fetch_script)
            )
        )
    )

    def fetched():
        try:
            fetched_file.remove()
            return True
        except Exception:
            return False

    r = runner.invoke(["discover"])
    assert not r.exception
    assert fetched()

    r = runner.invoke(["sync", "foobar"])
    assert not r.exception
    assert not fetched()

    r = runner.invoke(["sync"])
    assert not r.exception
    assert fetched()

    r = runner.invoke(["sync", "bambar"])
    assert not r.exception
    assert fetched()



import urllib.parse as urlparse

from .. import exceptions
from ..http import prepare_auth
from ..http import prepare_client_cert
from ..http import prepare_verify
from ..http import request
from ..http import USERAGENT
from ..vobject import Item
from ..vobject import split_collection
from .base import Storage


class HttpStorage(Storage):
    storage_name = "http"
    read_only = True
    _repr_attributes = ("username", "url")
    _items = None

    # Required for tests.
    _ignore_uids = True

    def __init__(
        self,
        url,
        username="",
        password="",
        verify=True,
        auth=None,
        useragent=USERAGENT,
        verify_fingerprint=None,
        auth_cert=None,
        **kwargs
    ):
        super().__init__(**kwargs)

        self._settings = {
            "auth": prepare_auth(auth, username, password),
            "cert": prepare_client_cert(auth_cert),
            "latin1_fallback": False,
        }
        self._settings.update(prepare_verify(verify, verify_fingerprint))

        self.username, self.password = username, password
        self.useragent = useragent

        collection = kwargs.get("collection")
        if collection is not None:
            url = urlparse.urljoin(url, collection)
        self.url = url
        self.parsed_url = urlparse.urlparse(self.url)

    def _default_headers(self):
        return {"User-Agent": self.useragent}

    def list(self):
        r = request("GET", self.url, headers=self._default_headers(), **self._settings)
        self._items = {}

        for item in split_collection(r.text):
            item = Item(item)
            if self._ignore_uids:
                item = item.with_uid(item.hash)

            self._items[item.ident] = item, item.hash

        return ((href, etag) for href, (item, etag) in self._items.items())

    def get(self, href):
        if self._items is None:
            self.list()

        try:
            return self._items[href]
        except KeyError:
            raise exceptions.NotFoundError(href)



import pytest
from requests import Response

from tests import normalize_item
from vdirsyncer.exceptions import UserError
from vdirsyncer.storage.http import HttpStorage
from vdirsyncer.storage.http import prepare_auth


def test_list(monkeypatch):
    collection_url = "http://127.0.0.1/calendar/collection.ics"

    items = [
        (
            "BEGIN:VEVENT\n"
            "SUMMARY:Eine Kurzinfo\n"
            "DESCRIPTION:Beschreibung des Termines\n"
            "END:VEVENT"
        ),
        (
            "BEGIN:VEVENT\n"
            "SUMMARY:Eine zweite Küèrzinfo\n"
            "DESCRIPTION:Beschreibung des anderen Termines\n"
            "BEGIN:VALARM\n"
            "ACTION:AUDIO\n"
            "TRIGGER:19980403T120000\n"
            "ATTACH;FMTTYPE=audio/basic:http://host.com/pub/ssbanner.aud\n"
            "REPEAT:4\n"
            "DURATION:PT1H\n"
            "END:VALARM\n"
            "END:VEVENT"
        ),
    ]

    responses = ["\n".join(["BEGIN:VCALENDAR"] + items + ["END:VCALENDAR"])] * 2

    def get(self, method, url, *a, **kw):
        assert method == "GET"
        assert url == collection_url
        r = Response()
        r.status_code = 200
        assert responses
        r._content = responses.pop().encode("utf-8")
        r.headers["Content-Type"] = "text/calendar"
        r.encoding = "ISO-8859-1"
        return r

    monkeypatch.setattr("requests.sessions.Session.request", get)

    s = HttpStorage(url=collection_url)

    found_items = {}

    for href, etag in s.list():
        item, etag2 = s.get(href)
        assert item.uid is not None
        assert etag2 == etag
        found_items[normalize_item(item)] = href

    expected = {
        normalize_item("BEGIN:VCALENDAR\n" + x + "\nEND:VCALENDAR") for x in items
    }

    assert set(found_items) == expected

    for href, etag in s.list():
        item, etag2 = s.get(href)
        assert item.uid is not None
        assert etag2 == etag
        assert found_items[normalize_item(item)] == href


def test_readonly_param():
    url = "http://example.com/"
    with pytest.raises(ValueError):
        HttpStorage(url=url, read_only=False)

    a = HttpStorage(url=url, read_only=True).read_only
    b = HttpStorage(url=url, read_only=None).read_only
    assert a is b is True


def test_prepare_auth():
    assert prepare_auth(None, "", "") is None

    assert prepare_auth(None, "user", "pwd") == ("user", "pwd")
    assert prepare_auth("basic", "user", "pwd") == ("user", "pwd")

    with pytest.raises(ValueError) as excinfo:
        assert prepare_auth("basic", "", "pwd")
    assert "you need to specify username and password" in str(excinfo.value).lower()

    from requests.auth import HTTPDigestAuth

    assert isinstance(prepare_auth("digest", "user", "pwd"), HTTPDigestAuth)

    with pytest.raises(ValueError) as excinfo:
        prepare_auth("ladida", "user", "pwd")

    assert "unknown authentication method" in str(excinfo.value).lower()


def test_prepare_auth_guess(monkeypatch):
    import requests_toolbelt.auth.guess

    assert isinstance(
        prepare_auth("guess", "user", "pwd"), requests_toolbelt.auth.guess.GuessAuth
    )

    monkeypatch.delattr(requests_toolbelt.auth.guess, "GuessAuth")

    with pytest.raises(UserError) as excinfo:
        prepare_auth("guess", "user", "pwd")

    assert "requests_toolbelt is too old" in str(excinfo.value).lower()


def test_verify_false_disallowed():
    with pytest.raises(ValueError) as excinfo:
        HttpStorage(url="http://example.com", verify=False)

    assert "forbidden" in str(excinfo.value).lower()
    assert "consider setting verify_fingerprint" in str(excinfo.value).lower()



"""
Vdirsyncer synchronizes calendars and contacts.

Please refer to https://vdirsyncer.pimutils.org/en/stable/packaging.html for
how to package vdirsyncer.
"""
from setuptools import Command
from setuptools import find_packages
from setuptools import setup


requirements = [
    # https://github.com/mitsuhiko/click/issues/200
    "click>=5.0,<9.0",
    "click-log>=0.3.0, <0.4.0",
    # https://github.com/pimutils/vdirsyncer/issues/478
    "click-threading>=0.5",
    "requests >=2.20.0",
    # https://github.com/sigmavirus24/requests-toolbelt/pull/28
    # And https://github.com/sigmavirus24/requests-toolbelt/issues/54
    "requests_toolbelt >=0.4.0",
    # https://github.com/untitaker/python-atomicwrites/commit/4d12f23227b6a944ab1d99c507a69fdbc7c9ed6d  # noqa
    "atomicwrites>=0.1.7",
]


class PrintRequirements(Command):
    description = "Prints minimal requirements"
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        for requirement in requirements:
            print(requirement.replace(">", "=").replace(" ", ""))


with open("README.rst") as f:
    long_description = f.read()


setup(
    # General metadata
    name="vdirsyncer",
    author="Markus Unterwaditzer",
    author_email="markus@unterwaditzer.net",
    url="https://github.com/pimutils/vdirsyncer",
    description="Synchronize calendars and contacts",
    license="BSD",
    long_description=long_description,
    # Runtime dependencies
    install_requires=requirements,
    # Optional dependencies
    extras_require={
        "google": ["requests-oauthlib"],
        "etesync": ["etesync==0.5.2", "django<2.0"],
    },
    # Build dependencies
    setup_requires=["setuptools_scm != 1.12.0"],
    # Other
    packages=find_packages(exclude=["tests.*", "tests"]),
    include_package_data=True,
    cmdclass={"minimal_requirements": PrintRequirements},
    use_scm_version={"write_to": "vdirsyncer/version.py"},
    entry_points={"console_scripts": ["vdirsyncer = vdirsyncer.cli:main"]},
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "License :: OSI Approved :: BSD License",
        "Operating System :: POSIX",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Internet",
        "Topic :: Utilities",
    ],
)



import pytest


class ServerMixin:
    @pytest.fixture
    def get_storage_args(
        self,
        request,
        tmpdir,
        slow_create_collection,
        radicale_server,
    ):
        def inner(collection="test"):
            url = "http://127.0.0.1:8001/"
            args = {
                "url": url,
                "username": "radicale",
                "password": "radicale",
            }

            if collection is not None:
                args = slow_create_collection(self.storage_class, args, collection)
            return args

        return inner
